import typing as T
from pydantic import BaseModel, parse_raw_as
from pydantic.fields import ModelField

ModelType = T.TypeVar("ModelType", bound=BaseModel)


def from_str(cls, v, field: ModelField) -> ModelType:
    if type(v) is str:
        res = parse_raw_as(field.outer_type_, v)
        return res
    return v


from enum import Enum

EnumType = T.TypeVar("EnumType", bound=Enum)


def transform_enum(
    original_value: T.Union[str, Enum],
    new_enum_type: T.Type[EnumType],
    ignore_null: bool = False,
) -> T.Optional[EnumType]:
    if ignore_null:
        if original_value is None:
            return None
    if isinstance(original_value, Enum):
        original_value = original_value.value
    new_value = (
        original_value.replace("/", "")
        .replace("-", "_")
        .replace("  ", " ")
        .replace(" ", "_")
    )
    return new_enum_type(new_value)


def enum_from_str(cls, v, field: ModelField) -> EnumType:
    if type(v) is str or not isinstance(v, field.type_):
        return transform_enum(original_value=v, new_enum_type=field.type_)
    return v



import typing as T
from uuid import UUID
from enum import Enum
import re
import edgedb
from devtools import debug
from pydantic import BaseModel, PrivateAttr
from .execute import query as execute_query
from .node import Node
from .constants import ALIAS_PATTERN, random_str
from .span import safe_span

NodeType = T.TypeVar("NodeType", bound=Node)

FILTER_FIELDS = ["_filter", "_limit", "_offset", "_order_by"]

ThisResolverType = T.TypeVar("ThisResolverType", bound="Resolver")


class ResolverException(Exception):
    pass


class UpdateOperation(str, Enum):
    """You can clear by choosing a resolver with REMOVE without filters"""

    REPLACE = ":="
    ADD = "+="
    REMOVE = "-="


EdgeResolverType = T.TypeVar("EdgeResolverType", bound="Resolver")


class DependencyInfo(BaseModel):
    fields: T.Set[str]
    ignore_if_admin: bool
    ignore_if_no_auth_id: bool
    require_user: bool


FieldDepMapAlias = T.Dict[str, DependencyInfo]


class FilterConnector(str, Enum):
    AND = " AND "
    OR = " OR "


from .nested_resolvers import NestedResolvers


class Resolver(BaseModel, T.Generic[NodeType]):
    """Can construct an EdgeQL query string from this!"""

    _node: T.ClassVar[T.Type[NodeType]]
    _edge_resolver_map: T.ClassVar[T.Dict[str, EdgeResolverType]]
    _field_dependency_map: T.ClassVar[FieldDepMapAlias]

    _filter: str = PrivateAttr(None)
    _order_by: str = PrivateAttr(None)
    _limit: int = PrivateAttr(None)
    _offset: int = PrivateAttr(None)
    _extra_fields: str = PrivateAttr(None)

    # this is for importing functions or libraries i think
    _modules: str = PrivateAttr(None)

    # this is just for updating
    _update_operation: UpdateOperation = PrivateAttr(None)

    _query_variables: T.Dict[str, T.Any] = PrivateAttr(default_factory=dict)
    _nested_resolvers: NestedResolvers = PrivateAttr(default_factory=NestedResolvers)

    _fields_to_return: T.Set[str] = PrivateAttr(None)
    _fields_to_include: T.Set[str] = PrivateAttr(default_factory=set)

    @staticmethod
    def replace_str_with_vars(s: str, d: dict) -> str:
        if not s:
            return ""
        for var_name, var_val in d.items():
            s = re.sub(
                rf"(\$)({var_name})(\W*)", r"\1" + f"[({var_val})]" + r"\3", s
            ).replace(f"[({var_val})]", f"{var_val}")
        return s

    def to_compare_filter_str(self) -> str:
        q_str = self.all_filters_str()
        q_str = self.replace_str_with_vars(s=q_str, d=self._query_variables)
        return q_str

    def to_compare_extra_str(self) -> str:
        return self.replace_str_with_vars(s=self._extra_fields, d=self._query_variables)

    def is_subset(self, other: "Resolver", should_debug: bool = False) -> bool:
        # first compare top level resolver, then go down from there
        # if self has more fields to return than other, self cannot be a subset
        if set(self.get_fields_to_return()) - set(other.get_fields_to_return()):
            if should_debug:
                print(f"self {self.__class__.__name__} has more fields than other")
            return False

        # if other does not have all extra fields that self has, return false
        if self._extra_fields:
            if self.to_compare_extra_str() not in other.to_compare_extra_str():
                return False

        # if self has different filters than other, self cannot be a subset
        if self.to_compare_filter_str() != other.to_compare_filter_str():
            if should_debug:
                print(
                    f"self {self.__class__.__name__} has diff filters, "
                    f"{self.to_compare_filter_str()=} {other.to_compare_filter_str()=}"
                )
            return False
        # now nested resolvers
        if not self._nested_resolvers.is_subset(other._nested_resolvers):
            return False
        return True

    """PROPERTIES"""

    @classmethod
    def set_field_dependency_map(cls, field_dep_map: FieldDepMapAlias) -> None:
        cls._field_dependency_map = field_dep_map

    @property
    def nested_resolvers(self) -> NestedResolvers:
        return self._nested_resolvers

    @classmethod
    @property
    def field_dependency_map(cls) -> FieldDepMapAlias:
        if getattr(cls, "_field_dependency_map", None) is None:
            cls.set_field_dependency_map(dict())
        return cls._field_dependency_map

    @classmethod
    def upsert_dependency_info(cls, field: str, dep_info: DependencyInfo) -> None:
        if getattr(cls, "_field_dependency_map", None) is None:
            cls.set_field_dependency_map(dict())
        prev_dep_info: DependencyInfo = cls._field_dependency_map.get(field)
        if prev_dep_info:
            prev_dep_info.fields.update(dep_info.fields)
            prev_dep_info.require_user = (
                prev_dep_info.require_user or dep_info.require_user
            )
            prev_dep_info.ignore_if_admin = (
                prev_dep_info.ignore_if_admin and dep_info.ignore_if_admin
            )
            prev_dep_info.ignore_if_no_auth_id = (
                prev_dep_info.ignore_if_no_auth_id and dep_info.ignore_if_no_auth_id
            )
        else:
            cls._field_dependency_map[field] = dep_info

    @classmethod
    @property
    def _model_name(cls) -> str:
        return cls._node._model_name

    def get_fields_to_return(self) -> T.List[str]:
        """Include nested fields with DOT notation"""
        if self._fields_to_return is None:
            fields = {
                *(self.get_aliases() - self._node._appendix_properties),
                *self._fields_to_include,
            }
        else:
            fields = self._fields_to_return
        return sorted(list(fields))

    @property
    def _filter_str(self) -> str:
        if not self._filter:
            return ""
        return f"FILTER {self._filter}"

    @property
    def _order_by_str(self) -> str:
        if not self._order_by:
            return ""
        return f"ORDER BY {self._order_by}"

    @property
    def _limit_str(self) -> str:
        if self._limit is None:
            return ""
        return f"LIMIT {self._limit}"

    @property
    def _offset_str(self) -> str:
        if self._offset is None:
            return ""
        return f"OFFSET {self._offset}"

    @property
    def filters_to_dict(self) -> T.Dict[str, T.Any]:
        return {
            "filter": self._filter,
            "order by": self._order_by,
            "limit": self._limit,
            "offset": self._offset,
        }

    @property
    def update_operation_str(self) -> str:
        if not self._update_operation:
            return ""
        return self._update_operation.value

    """SETTERS"""

    def fields_to_return(
        self: ThisResolverType, *fields_to_return: str
    ) -> ThisResolverType:
        self._fields_to_return = set(fields_to_return)
        return self

    def include_fields(
        self: ThisResolverType, *fields_to_include: str
    ) -> ThisResolverType:
        self._fields_to_include.update(set(fields_to_include))
        return self

    def exclude_fields(
        self: ThisResolverType, *fields_to_exclude: str
    ) -> ThisResolverType:
        self._fields_to_include = self._fields_to_include - set(fields_to_exclude)
        return self

    def include_computed_fields(self: ThisResolverType) -> ThisResolverType:
        self._fields_to_include.update(self._node._computed_properties)
        return self

    def include_appendix_fields(self: ThisResolverType) -> ThisResolverType:
        self._fields_to_include.update(self._node._appendix_properties)
        return self

    @staticmethod
    def validate_no_intersections_helper(
        set_a: T.Set = None, set_b: T.Set = None
    ) -> None:
        if set_a is None or set_b is None:
            return
        if used_keys := (set(set_a) & set(set_b)):
            raise ResolverException(
                f"Variable(s) {','.join(used_keys)} are already used."
            )

    def validate_against_query_variables(self, new_variables: dict = None) -> None:
        if not new_variables:
            return
        set_a = set(self._get_nested_query_variables().keys())
        set_b = set(new_variables.keys())
        self.validate_no_intersections_helper(set_a, set_b)

    def _add_variables(
        self, /, variables: T.Optional[T.Dict[str, T.Any]] = None
    ) -> None:
        # turn enums into strs
        if variables:
            variables = {
                k: v if not isinstance(v, Enum) else v.value
                for k, v in variables.items()
            }
        if variables is not None:
            self.validate_against_query_variables(variables)
            self._query_variables = {**self._query_variables, **variables}

    def modules(self: ThisResolverType, module_str: str) -> ThisResolverType:
        if self._modules:
            raise ResolverException(
                f"Module of {self._modules} has already been provided."
            )
        self._modules = module_str
        return self

    def filter(
        self: ThisResolverType,
        filter_str: str,
        variables: T.Optional[T.Dict[str, T.Any]] = None,
        connector: FilterConnector = FilterConnector.AND,
    ) -> ThisResolverType:
        if self._filter and not connector:
            raise ResolverException(
                f"Filter of {self._filter} has already been provided."
            )
        # TODO change variable names
        self._add_variables(variables)
        if not self._filter:
            self._filter = ""
        else:
            if connector and self._filter:
                self._filter += connector.value
        self._filter += filter_str
        return self

    def filter_by(self, connector: FilterConnector = FilterConnector.AND, **kwargs):
        """For now, only expecting one key value pair to be given, but could be more in future"""
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if not kwargs:
            raise ResolverException("Nothing to filter by.")
        conversion_map = self._node._edgedb_conversion_map
        filter_strs = []
        variables = {}
        for field_name, field_value in kwargs.items():
            cast = conversion_map[field_name]["cast"]
            variable_name = f"{field_name}{random_str(10)}"
            filter_strs.append(f".{field_name} = <{cast}>${variable_name}")
            variables[variable_name] = field_value
        filter_str = " AND ".join(filter_strs)
        return self.filter(
            filter_str=filter_str, variables=variables, connector=connector
        )

    def filter_in(
        self, connector: FilterConnector = FilterConnector.AND, **kwargs: list
    ):
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if not kwargs:
            raise ResolverException("Nothing to filter by.")
        conversion_map = self._node._edgedb_conversion_map
        filter_strs = []
        variables = {}
        for field_name, value_lst in kwargs.items():
            cast = conversion_map[field_name]["cast"]
            variable_name = f"{field_name}s{random_str(10)}"
            filter_strs.append(
                f".{field_name} in array_unpack(<array<{cast}>>${variable_name})"
            )
            variables[variable_name] = value_lst
        filter_str = " AND ".join(filter_strs)
        return self.filter(
            filter_str=filter_str, variables=variables, connector=connector
        )

    def extra_fields(
        self: ThisResolverType,
        extra_fields_str: str,
        variables: T.Optional[T.Dict[str, T.Any]] = None,
    ) -> ThisResolverType:
        if self._extra_fields:
            raise ResolverException(
                f"Extra fields of {self._extra_fields} has already been provided."
            )
        # TODO change variable names
        self._add_variables(variables)
        self._extra_fields = extra_fields_str
        return self

    def order_by(
        self: ThisResolverType,
        order_by_str: str,
        variables: T.Optional[T.Dict[str, T.Any]] = None,
    ) -> ThisResolverType:
        if self._order_by:
            raise ResolverException(
                f"Order by of {self._order_by} has already been provided."
            )
        self._add_variables(variables)
        self._order_by = order_by_str
        return self

    def limit(self: ThisResolverType, /, _: T.Optional[int]) -> ThisResolverType:
        if self._limit is not None:
            raise ResolverException(
                f"Limit of {self._limit} has already been provided."
            )
        self._limit = _
        return self

    def offset(self: ThisResolverType, /, _: T.Optional[int]) -> ThisResolverType:
        if self._offset is not None:
            raise ResolverException(
                f"Offset of {self._offset} has already been provided."
            )
        self._offset = _
        return self

    def update_operation(
        self: ThisResolverType,
        operation: UpdateOperation = None,
        *,
        add: bool = None,
        replace: bool = None,
        remove: bool = None,
    ) -> ThisResolverType:
        if self._update_operation:
            raise ResolverException(
                f"update_operation of {self._update_operation} has already been provided."
            )
        if operation:
            self._update_operation = operation
            return self
        if len([None for _ in [add, replace, remove] if _ is None]) != 2:
            raise ResolverException("Must give exactly one update operation")
        if add:
            opp = UpdateOperation.ADD
        elif replace:
            opp = UpdateOperation.REPLACE
        elif remove:
            opp = UpdateOperation.REMOVE
        else:
            raise ResolverException("Must provide an update operation")
        self._update_operation = opp
        return self

    """QUERY STRING"""

    def _get_nested_query_variables(self) -> T.Dict[str, T.Any]:
        query_variables = {**self._query_variables}
        for nested_resolver in self._nested_resolvers.all_resolvers():
            nested_query_variables = nested_resolver._get_nested_query_variables()
            if not nested_query_variables:
                continue
            if intersections := set(query_variables.keys()).intersection(
                set(nested_query_variables.keys())
            ):
                for intersection_key in intersections:
                    if (
                        query_variables[intersection_key]
                        != nested_query_variables[intersection_key]
                    ):
                        raise ResolverException(
                            f"'{intersection_key}' query value used multiple times with different value."
                        )
            query_variables = {**query_variables, **nested_query_variables}
        return query_variables

    def all_filters_str(self) -> str:
        s = ""
        if self._filter:
            s += f" {self._filter_str}"
        if self._order_by:
            s += f" {self._order_by_str}"
        if self._offset is not None:
            s += f" {self._offset_str}"
        if self._limit is not None:
            s += f" {self._limit_str}"
        return s

    def to_str(self, include_filters: bool = True, ignore_nested: bool = False) -> str:
        """
        OUTPUT: { name, created_at, bookings: { created_at } FILTER .created_at > $booking_created_at } FILTER .name = $name }
        """
        s = ""
        s += ", ".join(self.get_fields_to_return())
        if self._extra_fields:
            s += f", {self._extra_fields}"
        nested_resolver_str = self._nested_resolvers.to_str()
        if nested_resolver_str:
            s += f", {nested_resolver_str}"
        s = f"{{ {s} }}"
        if include_filters:
            s += self.all_filters_str()
        return s

    """HELPERS"""

    def has_filters(self) -> T.Optional[str]:
        for field in FILTER_FIELDS:
            if (val := getattr(self, field)) is not None:
                return val
        return None

    def clear_filters(self, remove_id_variable: bool = True) -> None:
        for field in FILTER_FIELDS:
            setattr(self, field, None)
        """Clearing filters does not clear query variables. 
        This is because you are only clearing the first level of filters."""
        if remove_id_variable:
            if "id" in self._query_variables:
                del self._query_variables["id"]

    """QUERY"""

    def full_query_str(self) -> str:
        module_str = "" if not self._modules else f"{self._modules} "
        return f"{module_str}SELECT {self._model_name} {self.to_str()}"

    def inner_delete_str(self) -> str:
        return f"DELETE {self._model_name} {self._filter_str}"

    def full_delete_str(self) -> str:
        return (
            f"WITH model := ({self.inner_delete_str()}) "
            f"SELECT model {self.to_str(include_filters=False)}"
        )

    async def query(
        self, given_client: edgedb.AsyncIOClient = None
    ) -> T.List[NodeType]:
        query_str = self.full_query_str()
        with safe_span(
            op=f"edgedb.query.{self._model_name}", description=query_str[:200]
        ):
            raw_d = await execute_query(
                client=given_client or self._node.db_client,
                query_str=query_str,
                variables=self._get_nested_query_variables(),
            )
        with safe_span(op=f"parse.{self._model_name}", description=f"{len(raw_d)=}"):
            return [self.parse_nested_obj(d) for d in raw_d]

    def clear_top_level_filters_and_variables(self) -> None:
        if self._filter:
            variables = re.findall(ALIAS_PATTERN, self._filter)
            for variable in variables:
                if variable in self._query_variables:
                    del self._query_variables[variable]
        self.clear_filters()

    async def get(
        self, given_client: edgedb.AsyncIOClient = None, **kwargs
    ) -> T.Optional[NodeType]:
        """Caveat is this must be a string value or an UUID"""
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if len(kwargs) != 1:
            raise ResolverException(f"Must only give one argument, received {kwargs}")
        # if getting, make sure there are no filters attached to the resolver
        if existing_filter_str := self.has_filters():
            raise ResolverException(
                f"This resolver already has filters: {existing_filter_str}. "
                f"If you wish to GET an object, use a new resolver."
            )
        key, value = list(kwargs.items())[0]
        if key not in self._node._exclusive_fields:
            raise ResolverException(f"Field '{key}' is not exclusive.")
        # this is to avoid nested collisions
        conversion_map = self._node._edgedb_conversion_map
        variables = {key: value}
        value_str = f'<{conversion_map[key]["cast"]}>${key}'
        self.filter(f".{key} = {value_str}", variables=variables)
        query_str = self.full_query_str()
        with safe_span(op=f"get.{self._model_name}", description=query_str[:200]):
            raw_d = await execute_query(
                client=given_client or self._node.db_client,
                query_str=query_str,
                variables=self._get_nested_query_variables(),
                only_one=True,
            )
        if not raw_d:
            return None
        with safe_span(op=f"parse.{self._model_name}"):
            return self.parse_nested_obj(raw_d)

    async def gerror(self, **kwargs) -> NodeType:
        model = await self.get(**kwargs)
        if not model:
            raise ResolverException(f"No {self._model_name} in db with fields {kwargs}")
        return model

    @classmethod
    def get_aliases(cls: T.Type[ThisResolverType]) -> T.Set[str]:
        return cls._node.get_aliases()

    def parse_nested_obj(
        self, raw_d: dict, error_for_extra_fields: bool = False
    ) -> NodeType:
        node: NodeType = self._node.parse_obj(raw_d)
        # pass on other fields as an "extra" dict
        # problem with this is aliases are currently also added to extra... do i want this?
        other_fields = set(raw_d.keys()) - self.get_aliases()
        for field_name in sorted(other_fields):
            resolver = self._nested_resolvers.resolver_from_field_name(field_name)
            if resolver is None:
                message = f"No nested resolver for {field_name=} found."
                if error_for_extra_fields:
                    raise ResolverException(message)
                node.extra[field_name] = raw_d[field_name]
                # print(message)
                continue
            nested_d = raw_d[field_name]
            value_to_save = nested_d
            if nested_d:
                if (list_or_set := type(nested_d)) in {list, set}:
                    val = list_or_set(resolver.parse_nested_obj(d) for d in nested_d)
                else:
                    val = resolver.parse_nested_obj(nested_d)
                value_to_save = val
            edge_name = NestedResolvers.edge_from_field_name(field_name)
            node.cache.add(
                edge=edge_name, resolver=resolver, val=value_to_save, raw_d=raw_d
            )
        node._used_resolver = self.copy(deep=True)
        node._original_dict = raw_d.copy()
        return node



import typing as T
from pydantic import BaseModel
from edgedb_orm import UNSET


class BasePatch(BaseModel):
    def updated_fields(
        self, exclude: T.Set[str] = None, exclude_last_updated: bool = False
    ) -> T.Set[str]:
        if exclude is None:
            exclude = set()
        if exclude_last_updated:
            exclude.add("last_updated")
        return {
            field
            for field in self.__fields__
            if field not in exclude and getattr(self, field) is not UNSET
        }



import typing as T
from pydantic import BaseModel

if T.TYPE_CHECKING:
    from .resolver import Resolver

SEPARATOR = "__AHH__"


class NestedResolvers(BaseModel):
    d: dict[str, list[T.Any]] = {}

    def get(self, edge: str) -> list["Resolver"]:
        return self.d.get(edge, [])

    def has(self, edge: str) -> bool:
        return edge in self.d

    def add(
        self,
        edge: str,
        resolver: "Resolver",
        ignore_if_subset: bool = False,
        make_first: bool = False,
    ) -> None:
        if not self.has(edge):
            self.d[edge] = []
        if ignore_if_subset:
            for r in list(self.get(edge)):
                if resolver.is_subset(r):
                    resolver = r
                    self.get(edge).remove(resolver)
                    break
        if make_first:
            self.d[edge].insert(0, resolver)
        else:
            self.d[edge].append(resolver)

    def has_subset(self, edge: str, resolver: "Resolver") -> bool:
        resolvers = self.get(edge)
        for r in resolvers:
            if resolver.is_subset(r, should_debug=True):
                return True
        return False

    def is_subset(self, other: "NestedResolvers") -> bool:
        # is this nested resolver a subset of the other
        for edge, resolvers in self.d.items():
            for r in resolvers:
                if not other.has_subset(edge, r):
                    return False
        return True

    def all_resolvers(self) -> list["Resolver"]:
        all_resolver_lst: list["Resolver"] = []
        for resolver_lst in self.d.values():
            all_resolver_lst.extend(resolver_lst)
        return all_resolver_lst

    def edge_to_str(self, edge: str) -> str:
        resolvers = self.get(edge)
        resolvers_str = []
        for i, r in enumerate(resolvers):
            if i == 0:
                resolvers_str.append(f"{edge}: {r.to_str()}")
            else:
                resolvers_str.append(
                    f"{edge}{SEPARATOR}{i} := (select .{edge} {r.to_str()})"
                )
        return ", ".join(resolvers_str)

    def to_str(self) -> str:
        return ", ".join([self.edge_to_str(e) for e in self.d.keys()])

    @staticmethod
    def edge_from_field_name(field_name: str) -> str:
        return field_name.split(SEPARATOR)[0]

    def resolver_from_field_name(self, field_name: str) -> T.Optional["Resolver"]:
        possible_edge = self.edge_from_field_name(field_name)
        resolvers = self.get(possible_edge)
        if not resolvers:
            return None
        if SEPARATOR not in field_name:
            return resolvers[0]
        try:
            index = int(field_name.split(SEPARATOR)[1])
            return resolvers[index]
        except (IndexError, ValueError) as e:
            print(f"{e=}, {field_name=}")
            return None



import typing as T
from devtools import debug
from pydantic import BaseModel

if T.TYPE_CHECKING:
    from .node import Node

    NodeType = T.TypeVar("NodeType", bound=Node)
    InsertType = T.TypeVar("InsertType", bound=BaseModel)


def mutation_strs_from_fields(
    field_names: T.List[str],
    conversion_map: T.Dict[str, T.Dict[str, str]],
    from_json: bool,
) -> T.List[str]:
    """[slug := <str>json_get(item, 'slug'), tags := <array<str>>json_get(item, 'tags')]"""
    field_strs: T.List[str] = []
    for field_name in field_names:
        if conversion_info := conversion_map.get(field_name):
            cast = f"<{conversion_info['cast']}>"
            if conversion_info["cardinality"] == "Many":
                cast = f"<array{cast}>"
            if from_json:
                value = f"{cast}json_get(item, '{field_name}')"
            else:
                value = f"{cast}${field_name}"
            if conversion_info["cardinality"] == "Many":
                value = f"array_unpack({value})"
            field_strs.append(f"{field_name} := {value}")
    return field_strs


def insert_str_from_cls(
    insert_cls: T.Type["InsertType"],
    node_cls: T.Type["NodeType"],
    edge_filter_strs: T.List[str],
    upsert_given_conflict_on: str = None,
    return_conflicting_model_on: str = None,
    custom_on_conflict_str: str = None,
    from_json: bool = True,
) -> str:
    """insert Person { first_name := <str>$first_name } OR insert Person { first_name := <str>item['first_name'] }"""
    insertable_fields = insert_cls.__fields__.keys()
    conversion_map = insert_cls._edgedb_conversion_map
    insertable_field_strs = mutation_strs_from_fields(
        field_names=insertable_fields,
        conversion_map=conversion_map,
        from_json=from_json,
    )
    insertable_field_strs.extend(edge_filter_strs)
    inner_insert_str = ", ".join(insertable_field_strs)
    # put this in a with block because of https://github.com/edgedb/edgedb/issues/3675
    edge_field_names = [s.strip().split(" ")[0] for s in edge_filter_strs]
    insertable_non_link_fields = set(insertable_fields) - set(
        node_cls._link_conversion_map.keys()
    )
    variable_fields = ", ".join(
        [f"{f} := {f}" for f in [*insertable_non_link_fields, *edge_field_names]]
    )
    # need to add edges to variable fields
    insert_str = (
        f"WITH {inner_insert_str} INSERT {node_cls._model_name} {{{variable_fields}}}"
    )
    conflict_str = ""
    if custom_on_conflict_str:
        conflict_str = custom_on_conflict_str
    elif upsert_given_conflict_on:
        fields_to_update = node_cls.GraphORM.updatable_fields & set(
            conversion_map.keys()
        )
        update_variable_fields = ", ".join(
            [f"{f} := {f}" for f in [*fields_to_update, *edge_field_names]]
        )
        conflict_str = (
            f"UNLESS CONFLICT ON .{upsert_given_conflict_on} ELSE "
            f"(UPDATE {node_cls._model_name} SET {{{update_variable_fields}}})"
        )
    elif return_conflicting_model_on:
        conflict_str = f"UNLESS CONFLICT ON .{return_conflicting_model_on} ELSE (SELECT {node_cls._model_name})"
        # conflict_str = f"UNLESS CONFLICT on .slug ELSE (SELECT {node_cls._model_name})"
    if conflict_str:
        insert_str += f" {conflict_str}"
    return insert_str


def update_str_from_cls() -> str:
    ...



import json
import re
import typing as T
from enum import Enum
import os
from pathlib import Path
from black import format_str, FileMode
import edgedb
from pydantic import BaseModel, parse_raw_as
from devtools import debug
from .introspection import (
    introspect_objects,
    introspect_scalars,
    ObjectType,
    ScalarType,
    Link,
    Property,
    Cardinality,
)
from .enums import PropertyCardinality


class DBVendor(str, Enum):
    edgedb = "edgedb"


class TLSSecurity(str, Enum):
    insecure = "insecure"


class PropertyConfig(BaseModel):
    module_path: str
    module_name: str
    validate_as_basemodel: bool = True
    cardinality: PropertyCardinality = PropertyCardinality.ONE


class NodeConfig(BaseModel):
    module_path: str = None
    appendix_properties: T.List[str] = []
    basemodel_properties: T.Dict[str, PropertyConfig] = {}


class DBConfig(BaseModel):
    vendor: DBVendor
    host: str
    password: str
    port: int = 5656
    tls_security: TLSSecurity
    copy_config: str = None
    hydrate: bool = False
    nodes: T.Dict[str, NodeConfig] = dict()


PATH_TO_MODULE = "edgedb_orm"
DEFAULT_INDENT = "    "
CONFIG_NAME = "GraphORM"


class GeneratorException(Exception):
    pass


def indent_lines(s: str, indent: str = DEFAULT_INDENT) -> str:
    chunks = s.split("\n")
    return indent + f"\n{indent}".join(chunks)


def imports(enums_module: str) -> str:
    lines = [
        "from __future__ import annotations",
        "import typing as T",
        "from enum import Enum",
        "from datetime import datetime, date, timedelta",
        "from uuid import UUID",
        "from decimal import Decimal",
        "from edgedb import RelativeDuration, AsyncIOClient, create_async_client",
        "from pydantic import BaseModel, Field, PrivateAttr, validator",
        f"from {PATH_TO_MODULE} import Node, Resolver, NodeException, ResolverException, UpdateOperation, Batch, UNSET, UnsetType, ComputedPropertyException, AppendixPropertyException, from_str, enum_from_str, BasePatch, FilterConnector",
        f"from .{enums_module} import *",
        'EdgeResolverType = T.TypeVar("EdgeResolverType", bound=Resolver)',
    ]
    return "\n".join(lines)


def build_enum_imports() -> str:
    lines = ["from enum import Enum"]
    return "\n".join(lines)


async def build_enums(client: edgedb.AsyncIOClient, include_strawberry: bool) -> str:
    scalar_types = await introspect_scalars(client)
    enum_strs: T.List[str] = []
    enum_names: T.List[str] = []
    for scalar in scalar_types:
        if not scalar.enum_values:
            continue
        enum_value_strs: T.List[str] = [f'{e} = "{e}"' for e in scalar.enum_values]
        enum_value_str = "\n".join(enum_value_strs)
        s = f"class {scalar.node_name}(str, Enum):\n{indent_lines(enum_value_str)}"
        enum_strs.append(s)
        enum_names.append(scalar.node_name)
    if include_strawberry:
        straw_enums_str = "\n".join(
            [f"{enum_name} = strawberry.enum({enum_name})" for enum_name in enum_names]
        )
        straw_lines = [
            "import strawberry",
            f"class Strawberry:\n{indent_lines(straw_enums_str)}",
        ]
        enum_strs.extend(straw_lines)
    return "\n".join(enum_strs)


def build_node_link_function_str(link: Link) -> str:
    link_resolver_name = f"{link.target.model_name}Resolver"
    return f"""
async def {link.name}(
    self,
    resolver: {link_resolver_name} = None,
    refresh: bool = False,
    revert_to_first: bool = False,
) -> {link.type_str}:
    return await self.resolve(
        edge_name="{link.name}",
        edge_resolver=resolver or {link_resolver_name}(),
        refresh=refresh,
        revert_to_first=revert_to_first,
    )
    """


def build_resolver_link_function_str(node_resolver_name: str, link: Link) -> str:
    link_resolver_name = f"{link.target.model_name}Resolver"
    return f"""
def {link.name}(self, _: T.Optional[{link_resolver_name}] = None, /, ignore_if_subset: bool = False, make_first: bool = False) -> {node_resolver_name}:
    self._nested_resolvers.add("{link.name}", _ or {link_resolver_name}(), ignore_if_subset=ignore_if_subset, make_first=make_first)
    return self
    """


def build_get_functions_str(node_name: str, exclusive_field_names: T.Set[str]) -> str:
    exclusive_field_names = sorted(list(exclusive_field_names))
    params_fields_str = ", ".join(
        [f"{f}: T.Optional[T.Any] = None" for f in exclusive_field_names]
    )
    dict_fields_str = ", ".join([f'"{f}": {f}' for f in exclusive_field_names])
    get_str = f"""
async def get(self, given_client: AsyncIOClient = None, {params_fields_str}) -> T.Optional[{node_name}]:
    return await super().get(given_client=given_client, **{{{dict_fields_str}}})
    """
    gerror_str = f"""
async def gerror(self, given_client: AsyncIOClient = None, {params_fields_str}) -> {node_name}:
    return await super().gerror(given_client=given_client, **{{{dict_fields_str}}})
    """
    return f"{get_str}\n{gerror_str}"


def build_include_fields_function(
    node_resolver_name: str,
    appendix_properties: set[str],
    computed_properties: set[str],
) -> str:
    field_names = sorted({*appendix_properties, *computed_properties})
    names_params = ", ".join([f"{name}: bool = False" for name in field_names])
    logic_strs = "\n".join(
        [
            f'if {field_name} is True: fields_to_include.add("{field_name}")'
            for field_name in field_names
        ]
    )
    names_params_str = "" if not names_params else f", {names_params}"
    s = f"""
def include_fields(self, *other_fields: str{names_params_str}) -> {node_resolver_name}:
    fields_to_include: set[str] = set(other_fields)
{indent_lines(logic_strs)}
    return super().include_fields(*fields_to_include)
    """
    return s


def build_filter_functions_str(node_name: str, conversion_map: T.Dict[str, dict]):
    field_names = sorted(conversion_map.keys())
    params_fields_str = ", ".join(
        [f"{f}: T.Optional[T.Any] = None" for f in field_names]
    )
    dict_fields_str = ", ".join([f'"{f}": {f}' for f in field_names])
    filter_by_str = f"""
def filter_by(self, filter_connector: FilterConnector = FilterConnector.AND, {params_fields_str}) -> {node_name}Resolver:
    return super().filter_by(connector=filter_connector, **{{{dict_fields_str}}})
    """
    # now for filter in
    params_fields_lst_str = ", ".join(
        [f"{f}: T.Optional[T.List[T.Any]] = None" for f in field_names]
    )
    filter_in_str = f"""
def filter_in(self, filter_connector: FilterConnector = FilterConnector.AND, {params_fields_lst_str}) -> {node_name}Resolver:
    return super().filter_in(connector=filter_connector, **{{{dict_fields_str}}})
    """
    return f"{filter_by_str}\n{filter_in_str}"


def build_update_function_str(node_resolver_name: str, links: T.List[Link]) -> str:
    link_strs: T.List[str] = []
    link_names: T.List[str] = []
    for link in links:
        if link.readonly:
            continue
        link_strs.append(
            f"{link.name}: T.Optional[{link.target.model_name}Resolver] = None"
        )
        link_names.append(link.name)
    link_params_str = ", ".join(link_strs)
    return f"""
async def update(
    self,
    given_resolver: {node_resolver_name} = None,
    error_if_no_update: bool = False,
    batch: Batch = None,
    given_client: AsyncIOClient = None,
    {link_params_str}
) -> None:
    set_links_d = {{{", ".join([f'"{link_name}": {link_name}' for link_name in link_names])}}}
    set_links_d = {{key: val for key, val in set_links_d.items() if val is not None}}

    return await super().update(
        given_resolver=given_resolver,
        error_if_no_update=error_if_no_update,
        set_links_d=set_links_d,
        batch=batch,
        given_client=given_client
    )
    """


def add_quotes(lst: T.Iterable[str]) -> T.Iterable[str]:
    return [f'"{o}"' for o in lst]


def build_orm_config(
    model_name: str, updatable_fields: T.Set[str], exclusive_fields: T.Set[str]
) -> str:
    return f"""
class {CONFIG_NAME}:
    model_name = "{model_name}"
    client = client
    updatable_fields: T.Set[str] = {{{', '.join(add_quotes(sorted(list(updatable_fields))))}}}
    exclusive_fields: T.Set[str] = {{{', '.join(add_quotes(sorted(list(exclusive_fields))))}}}
    """


def stringify_dict(
    d: T.Union[T.Dict[str, str], str], stringify_value: bool = True
) -> str:
    if type(d) is not dict:
        s = f"{d}"
        if type(d) is not bool:
            if stringify_value:
                s = f'"{s}"'
        return s
    inner = [
        f'"{k}":{stringify_dict(v, stringify_value=stringify_value)}'
        for k, v in d.items()
    ]
    return f"{{{','.join(inner)}}}"


def stringify_set(s: T.Set[str]) -> str:
    if not s:
        return "set()"
    s_sorted = sorted(list(s))
    strs: T.List[str] = [f'"{i}"' for i in s_sorted]
    return "{" + ",".join(strs) + "}"


def edgedb_conversion_type_from_prop(prop: Property) -> str:
    """
    s = prop.target.name
    pattern = r"default::\w+"
    s = re.sub(pattern, "std::str", s)
    return s
    """
    return prop.target.name


def build_validator_module_imports(db_config: DBConfig) -> str:
    if not db_config.nodes:
        return ""
    node_configs = db_config.nodes.values()
    from_str_import_strs: T.List[str] = []
    for node_config in node_configs:
        for prop_name, prop_config in node_config.basemodel_properties.items():
            from_str_import_strs.append(
                f"from {prop_config.module_path} import {prop_config.module_name}"
            )
    return "\n".join(sorted(list(set(from_str_import_strs))))


def build_from_str_validator_str(node_config: T.Optional[NodeConfig]) -> str:
    if not node_config:
        return ""
    field_name_strs: T.List[str] = []
    for field_name, props_config in node_config.basemodel_properties.items():
        if props_config.validate_as_basemodel is False:
            continue
        key = field_name
        if field_name in node_config.appendix_properties:
            key += "_"
        field_name_strs.append(f'"{key}"')
    if not field_name_strs:
        return ""
    return f"""
_from_str = validator({", ".join(field_name_strs)}, pre=True, allow_reuse=True)(from_str)
    """


def remove_falsies(lst: list) -> list:
    return [i for i in lst if i]


def build_node_and_resolver(
    object_type: ObjectType,
    node_config: T.Optional[NodeConfig],
    edge_resolver_map_strs: T.List[str],
    hydrate: bool,
    dehydrate: bool,
) -> str:
    # need to sort props and links by required, exclusive, no default, rest
    object_type.properties.sort(
        key=lambda x: f"{not x.is_computed}-{x.required}-{x.is_exclusive}-{x.default}",
        reverse=True,
    )
    object_type.links.sort(
        key=lambda x: f"{not x.is_computed}-{x.required}-{x.is_exclusive}-{x.default}",
        reverse=True,
    )
    # start with the properties
    node_resolver_name = f"{object_type.node_name}Resolver"
    property_strs: T.List[str] = []
    insert_property_strs: T.List[str] = []
    patch_property_strs: T.List[str] = []
    updatable_fields: T.Set[str] = set()
    exclusive_fields: T.Set[str] = set()

    node_edgedb_conversion_map: T.Dict[str, T.Dict[str, str]] = {}
    insert_edgedb_conversion_map: T.Dict[str, T.Dict[str, str]] = {}
    patch_edgedb_conversion_map: T.Dict[str, T.Dict[str, str]] = {}

    computed_properties: T.Set[str] = set()
    computed_property_getter_strs: T.List[str] = []

    appendix_properties: T.Set[str] = set()

    for prop in object_type.properties:
        conversion_type = edgedb_conversion_type_from_prop(prop)
        node_edgedb_conversion_map[prop.name] = {
            "cast": conversion_type,
            "cardinality": prop.cardinality.value,
            "readonly": prop.readonly,
        }
        if prop.is_computed:
            computed_properties.add(prop.name)
        is_appendix = False
        if node_config:
            if prop.name in node_config.appendix_properties:
                is_appendix = True
                appendix_properties.add(prop.name)
        if not prop.readonly and not prop.is_computed:
            updatable_fields.add(prop.name)
        if prop.is_exclusive:
            exclusive_fields.add(prop.name)
        default_value_str = "..." if prop.required else "None"
        allow_mutation_str = (
            f"allow_mutation={not prop.readonly and not prop.is_computed}"
        )
        if node_config and prop.name in node_config.basemodel_properties:
            prop_config = node_config.basemodel_properties[prop.name]
            module_name = prop_config.module_name
            type_str = prop.type_str_basemodel(
                module_name, cardinality=prop_config.cardinality
            )
        else:
            type_str = prop.type_str
        if not (prop.is_computed or is_appendix):
            property_strs.append(
                f"{prop.name}: {type_str} = Field({default_value_str}, {allow_mutation_str})"
            )
        else:
            if prop.is_computed:
                exception_name = "ComputedPropertyException"
                property_name = f"_{prop.name}"
                property_str = f"{property_name}: T.Union[{type_str}, UnsetType] = PrivateAttr(default_factory=UnsetType)"
            else:
                exception_name = "AppendixPropertyException"
                property_name = f"{prop.name}_"
                property_str = f'{property_name}: T.Union[{type_str}, UnsetType] = Field(default_factory=UnsetType, alias="{prop.name}")'
            property_strs.append(property_str)
            val_str = "val"
            if "Set[" in type_str:
                val_str = f"val if type(val) != list else set(val)"
            computed_property_getter_strs.append(
                f"""
@property
def {prop.name}(self) -> {type_str}:
    if self.{property_name} is UNSET:
        if "{prop.name}" in self.extra:
            val = self.extra["{prop.name}"]
            self.{property_name} = {val_str}
        else:
            raise {exception_name}("{prop.name} is unset")
    return self.{property_name}
                """
            )
            if not prop.is_computed:
                computed_property_getter_strs.append(
                    f"""
@{prop.name}.setter
def {prop.name}(self, {prop.name}: {type_str}) -> None:
    self.{property_name} = {prop.name}
                    """
                )
        if prop.name != "id":
            # for insert type
            if not prop.is_computed and not prop.not_insertable:
                insert_edgedb_conversion_map[prop.name] = {
                    "cast": conversion_type,
                    "cardinality": prop.cardinality.value,
                    "readonly": prop.readonly,
                }
                insert_type_str = type_str
                # if required but has default, add optional back
                if prop.required and prop.default:
                    insert_type_str = f"T.Optional[{insert_type_str}]"
                default_value_str = (
                    " = None" if insert_type_str.startswith("T.Optional[") else ""
                )
                insert_property_strs.append(
                    f"{prop.name}: {insert_type_str}{default_value_str}"
                )
            # for update type
            if not prop.is_computed and not prop.readonly:
                patch_edgedb_conversion_map[prop.name] = {
                    "cast": conversion_type,
                    "cardinality": prop.cardinality.value,
                    "readonly": prop.readonly,
                }
                patch_type_str = type_str
                if prop.required and prop.default:
                    patch_type_str = f"T.Optional[{patch_type_str}]"
                final_patch_type_str = f"T.Union[{patch_type_str}, UnsetType]"
                patch_property_strs.append(
                    f"{prop.name}: {final_patch_type_str} = Field(default_factory=UnsetType)"
                )

    link_function_strs: T.List[str] = []
    resolver_function_strs: T.List[str] = []
    updatable_links: T.Set[str] = set()
    exclusive_links: T.Set[str] = set()
    link_conversion_map: T.Dict[str, T.Dict[str, str]] = {}
    edge_resolver_map: T.Dict[str, str] = {}

    for link in object_type.links:
        if link.name == "__type__":
            continue
        link_conversion_map[link.name] = {
            "cast": link.target.model_name,
            "cardinality": link.cardinality.value,
            "readonly": link.readonly,
            "required": link.required,
        }
        edge_resolver_map[link.name] = f"{link.target.model_name}Resolver"
        if not link.readonly and not link.is_computed:
            updatable_links.add(link.name)
        if link.is_exclusive:
            exclusive_links.add(link.name)
        link_function_strs.append(build_node_link_function_str(link))
        resolver_function_strs.append(
            build_resolver_link_function_str(
                node_resolver_name=node_resolver_name, link=link
            )
        )
        # for insert
        if not link.is_computed and not link.not_insertable:
            insert_resolver_str = f"{link.target.model_name}Resolver"
            if (not link.required) or (link.required and link.default):
                insert_resolver_str = f"T.Optional[{insert_resolver_str}]"
            default_value_str = (
                " = None" if insert_resolver_str.startswith("T.Optional[") else ""
            )
            insert_property_strs.append(
                f"{link.name}: {insert_resolver_str}{default_value_str}"
            )
        # for update
        if not link.is_computed and not link.readonly:
            patch_resolver_str = f"{link.target.model_name}Resolver"
            if (not link.required) or (link.required and link.default):
                patch_resolver_str = f"T.Optional[{patch_resolver_str}]"
            final_patch_resolver_str = f"T.Union[{patch_resolver_str}, UnsetType]"
            patch_property_strs.append(
                f"{link.name}: {final_patch_resolver_str} = Field(default_factory=UnsetType)"
            )

    orm_config_str = build_orm_config(
        model_name=object_type.node_name,
        updatable_fields={*updatable_fields, *updatable_links},
        exclusive_fields={*exclusive_fields, *exclusive_links},
    )

    insert_model_name = f"{object_type.node_name}Insert"
    patch_model_name = f"{object_type.node_name}Patch"

    # insert type
    insert_inner_str = "\n".join(insert_property_strs)
    insert_conversion_map_str = f"_edgedb_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(insert_edgedb_conversion_map)}"
    insert_s = f"class {insert_model_name}(BaseModel):\n{indent_lines(insert_inner_str)}\n\n{indent_lines(insert_conversion_map_str)}"

    # patch type
    patch_inner_str = "\n".join(patch_property_strs)
    patch_conversion_map_str = f"_edgedb_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(patch_edgedb_conversion_map)}"
    patch_s = f"class {patch_model_name}(BasePatch):\n{indent_lines(patch_inner_str)}\n\n{indent_lines(patch_conversion_map_str)}"

    # node
    node_properties_str = "\n".join(property_strs)
    from_str_validator_str = build_from_str_validator_str(node_config=node_config)
    if hydrate:
        node_properties_str = ""
        from_str_validator_str = ""

    computed_property_getter_str = "\n".join(computed_property_getter_strs)
    node_conversion_map_str = f"_edgedb_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(node_edgedb_conversion_map)}"
    insert_link_conversion_map_str = f"_link_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(link_conversion_map)}"
    computed_properties_str = f"_computed_properties: T.ClassVar[T.Set[str]] = {stringify_set(computed_properties)}"
    appendix_properties_str = f"_appendix_properties: T.ClassVar[T.Set[str]] = {stringify_set(appendix_properties)}"
    basemodel_properties = (
        [] if not node_config else node_config.basemodel_properties.keys()
    )
    basemodel_properties_str = f"_basemodel_properties: T.ClassVar[T.Set[str]] = {stringify_set(set(basemodel_properties))}"
    node_link_functions_str = "\n".join(link_function_strs)
    update_function_str = build_update_function_str(
        node_resolver_name=node_resolver_name, links=object_type.links
    )
    if dehydrate:
        node_link_functions_str = ""
        update_function_str = ""
    node_inner_strs = [
        node_properties_str,
        "\n",
        from_str_validator_str,
        "\n",
        computed_property_getter_str,
        node_conversion_map_str,
        insert_link_conversion_map_str,
        computed_properties_str,
        appendix_properties_str,
        basemodel_properties_str,
        node_link_functions_str,
        update_function_str,
        orm_config_str,
    ]
    node_inner_str = "\n".join(remove_falsies(node_inner_strs))
    node_outer_strs = []
    node_outer_str = "\n".join(remove_falsies(node_outer_strs))
    inherits = (
        f"Node[{insert_model_name}, {patch_model_name}]"
        if not hydrate
        else f"{object_type.node_name}Hydrated"
    )
    node_s = f"{node_outer_str}\nclass {object_type.node_name}({inherits}):\n{indent_lines(node_inner_str)}"

    # resolver
    resolver_properties_str = f"_node = {object_type.node_name}"
    resolver_link_functions_str = "\n".join(resolver_function_strs)
    resolver_get_functions_str = build_get_functions_str(
        node_name=object_type.node_name, exclusive_field_names=exclusive_fields
    )
    resolver_include_fields_str = build_include_fields_function(
        node_resolver_name=node_resolver_name,
        appendix_properties=appendix_properties,
        computed_properties=computed_properties,
    )
    resolver_filter_functions_str = build_filter_functions_str(
        node_name=object_type.node_name, conversion_map=node_edgedb_conversion_map
    )

    edge_resolver_val_strs = [
        f"T.Type[{v}]" for v in sorted(set(edge_resolver_map.values()))
    ]
    union_type = (
        f"T.Union[{', '.join(edge_resolver_val_strs)}]"
        if edge_resolver_map
        else "EdgeResolverType"
    )
    edge_resolver_map_strs.append(
        f"{node_resolver_name}._edge_resolver_map: T.Dict[str, {union_type}] ="
        f" {stringify_dict(edge_resolver_map, stringify_value=False)}"
    )
    resolver_inner_strs = [
        resolver_properties_str,
        resolver_link_functions_str,
        resolver_get_functions_str,
        resolver_filter_functions_str,
        resolver_include_fields_str,
    ]
    resolver_inner_str = "\n".join(resolver_inner_strs)
    resolver_s = f"class {node_resolver_name}(Resolver[{object_type.node_name}]):\n{indent_lines(resolver_inner_str)}"

    final_s = (
        f"{object_type.node_name}.{CONFIG_NAME}.resolver_type = {node_resolver_name}"
    )
    return f"{insert_s}\n{patch_s}\n{node_s}\n{resolver_s}\n{final_s}"


async def build_nodes_and_resolvers(
    client: edgedb.AsyncIOClient,
    db_config: DBConfig,
    nodes_to_hydrate: T.Set[str],
    dehydrate: bool,
) -> str:
    object_types = await introspect_objects(client)
    node_strs: T.List[str] = []
    edge_resolver_map_strs: T.List[str] = []
    for object_type in object_types:
        node_strs.append(
            build_node_and_resolver(
                object_type,
                node_config=db_config.nodes.get(object_type.node_name),
                edge_resolver_map_strs=edge_resolver_map_strs,
                hydrate=object_type.node_name in nodes_to_hydrate and not dehydrate,
                dehydrate=dehydrate,
            )
        )
    update_forward_refs_inserts_str = "\n".join(
        [f"{o.node_name}Insert.update_forward_refs()" for o in object_types]
    )
    update_forward_refs_nodes_str = "\n".join(
        [f"{o.node_name}.update_forward_refs()" for o in object_types]
    )
    nodes_str = "\n".join(node_strs)
    edge_resolver_map_str = "\n".join(edge_resolver_map_strs)
    return f"{nodes_str}\n\n{update_forward_refs_inserts_str}\n\n{update_forward_refs_nodes_str}\n\n{edge_resolver_map_str}"


def build_client(db_config: DBConfig) -> str:
    return f"""
client = create_async_client(
    tls_security="{db_config.tls_security.value}",
    host="{db_config.host}",
    password="{db_config.password}", 
    port={db_config.port}
)
    """


def validate_output_path(path: Path) -> None:
    if not os.path.isdir(path):
        if os.path.isfile(path):
            raise GeneratorException(
                f"output path {path=} must be a directory, not a file."
            )
        if not os.path.exists(path):
            os.makedirs(path)


def build_hydrate_imports(db_config: DBConfig) -> str:
    import_strs: T.List[str] = []
    for node_name, config in db_config.nodes.items():
        if config.module_path:
            hydrated_name = f"{node_name}Hydrated"
            import_strs.append(
                f"from {config.module_path} import {node_name} as {hydrated_name}"
            )
            # idk why i have the part below, was just copying from dgraph_orm, seems unecessary and wrong
            # import_strs.append(
            #     f"{hydrated_name}.{CONFIG_NAME}.resolver._node = {hydrated_name}"
            # )
    return "\n".join(import_strs)


def get_nodes_to_hydrate(db_config: DBConfig) -> T.Set[str]:
    node_names: T.Set[str] = set()
    for node_name, config in db_config.nodes.items():
        if config.module_path:
            node_names.add(node_name)
    return node_names


async def build_enums_from_config(db_config: DBConfig, include_strawberry: bool) -> str:
    client = edgedb.create_async_client(
        tls_security=db_config.tls_security,
        host=db_config.host,
        password=db_config.password,
        port=db_config.port,
    )
    enums_imports = build_enum_imports()
    enums_str = await build_enums(client, include_strawberry=include_strawberry)
    s = "\n".join([enums_imports, enums_str])
    s = format_str(s, mode=FileMode())
    return s


async def build_from_config(
    db_config: DBConfig,
    enums_module: str,
    hydrate: bool = False,
    dehydrate: bool = False,
) -> str:
    client = edgedb.create_async_client(
        tls_security=db_config.tls_security,
        host=db_config.host,
        password=db_config.password,
        port=db_config.port,
    )
    imports_str = imports(enums_module=enums_module)
    client_str = build_client(db_config)
    validator_module_imports = build_validator_module_imports(db_config)
    hydrate_imports = "" if not hydrate else build_hydrate_imports(db_config)
    nodes_and_resolvers_str = await build_nodes_and_resolvers(
        client,
        db_config=db_config,
        nodes_to_hydrate=get_nodes_to_hydrate(db_config),
        dehydrate=dehydrate,
    )

    s = "\n".join(
        [
            imports_str,
            client_str,
            validator_module_imports,
            hydrate_imports,
            nodes_and_resolvers_str,
        ]
    )
    s = format_str(s, mode=FileMode())
    return s


async def generate(
    config_path: Path, output_path: Path, include_strawberry: bool = False
) -> None:
    validate_output_path(output_path)
    all_config_json_str = open(config_path).read()
    all_config_d = parse_raw_as(T.Dict[str, DBConfig], all_config_json_str)
    for db_name, db_config in all_config_d.items():
        if db_config.copy_config:
            db_config_to_copy = all_config_d[db_config.copy_config]
            db_config.hydrate = db_config_to_copy.hydrate
            db_config.nodes = db_config_to_copy.nodes

        # first build enums folder
        enums_s = await build_enums_from_config(
            db_config=db_config, include_strawberry=include_strawberry
        )
        enums_module = f"{db_name}_enums"
        open(output_path / f"{enums_module}.py", "w").write(enums_s)
        hydrate = db_config.hydrate
        # must include strawberry types in both for circular dependency reasons
        s = await build_from_config(
            db_config=db_config,
            dehydrate=hydrate,
            enums_module=enums_module,
        )
        open(output_path / f"{db_name}.py", "w").write(s)
        if hydrate:
            s = await build_from_config(
                db_config=db_config,
                hydrate=True,
                enums_module=enums_module,
            )
            open(output_path / f"{db_name}_hydrated.py", "w").write(s)



import typing as T
import time
from pydantic import BaseModel
from .unset import UNSET, UnsetType

if T.TYPE_CHECKING:
    from .resolver import Resolver


class CacheNode(BaseModel):
    val: T.Any
    resolver: T.Any  # circular unfortunately
    raw_d: dict
    timestamp: float


class Cache(BaseModel):
    d: dict[str, list[CacheNode]] = {}

    def get(self, edge: str) -> list[CacheNode]:
        return self.d.get(edge, [])

    def has(self, edge: str) -> bool:
        return edge in self.d

    def add(self, edge: str, resolver: "Resolver", val: T.Any, raw_d: dict) -> None:
        if not self.has(edge):
            self.d[edge] = []
        self.d[edge].append(
            CacheNode(resolver=resolver, val=val, raw_d=raw_d, timestamp=time.time())
        )

    def replace(self, edge: str, cache_nodes: list[CacheNode]) -> None:
        self.d[edge] = cache_nodes

    def clear(self, edge: str) -> None:
        if edge in self.d:
            del self.d[edge]

    def remove(self, edge: str, resolver: "Resolver") -> None:
        nodes = self.get(edge)
        new_nodes = []
        for node in list(nodes):
            if node.resolver is not resolver:
                new_nodes.append(node.resolver)
        self.d[edge] = new_nodes

    def val(
        self, edge: str, resolver: "Resolver", revert_to_first: bool = False
    ) -> T.Union[T.Any, UnsetType]:
        nodes = self.get(edge)
        if not nodes:
            return UNSET
        first_superset_node: T.Optional[CacheNode] = None
        for node in nodes:
            if node.resolver is resolver:
                return node.val
            if not first_superset_node:
                if resolver.is_subset(node.resolver, should_debug=True):
                    first_superset_node = node
        if first_superset_node:
            return first_superset_node.val
        if revert_to_first:
            return nodes[0].val
        return UNSET

    def is_empty(self) -> bool:
        return len(self.d) == 0



from pydantic import BaseModel
import typing as T


class UnsetType(BaseModel):
    __instance: T.Optional["UnsetType"] = None

    def __new__(cls: T.Type["UnsetType"]) -> "UnsetType":
        if cls.__instance is None:
            ret = super().__new__(cls)
            cls.__instance = ret
            return ret
        else:
            return cls.__instance

    def __str__(self):
        return ""

    def __repr__(self) -> str:
        return "UNSET"

    def __bool__(self):
        return False


UNSET: T.Any = UnsetType()


class ComputedPropertyException(Exception):
    pass


class AppendixPropertyException(Exception):
    pass



import time
from contextlib import contextmanager
import os

PRINT_SPAN = bool(int(os.getenv("_PRINT_SPAN", "0")))

try:
    from sentry_sdk import start_span

    USE_SENTRY = True
except Exception:
    print("could not import sentry_sdk for span")
    USE_SENTRY = False


@contextmanager
def print_span(op: str, description: str = None):
    start = time.time()
    yield
    time_took = time.time() - start
    if PRINT_SPAN:
        print(f"{op=}, {description=} took {time_took*1_000:.2f} ms.")


@contextmanager
def safe_span(op: str, description: str = None, use: bool = True, **kwargs):
    if not USE_SENTRY or not use:
        with print_span(op=op, description=description):
            yield
    else:
        if USE_SENTRY is True:
            with start_span(op=op, description=description, **kwargs):
                yield



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edgedb_orm']

package_data = \
{'': ['*']}

install_requires = \
['black',
 'devtools',
 'edgedb>=0.23.0,<0.24.0',
 'fastapi',
 'orjson>=3.6,<4.0',
 'pydantic[email]']

setup_kwargs = {
    'name': 'edgedb-orm',
    'version': '0.2.7',
    'description': '',
    'long_description': None,
    'author': 'Jeremy Berman',
    'author_email': 'jerber@sas.upenn.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import re
import typing as T
import time
import orjson
from devtools import debug
import edgedb
from .span import safe_span
from .constants import RE_CODE

MUTATION_ACTIONS = ["insert ", "update ", "delete "]


class ExecuteException(Exception):
    pass


def operation_from_query_str(query_str: str) -> str:
    s = query_str.lower()
    for action in MUTATION_ACTIONS:
        if action in s:
            return "mutation"
    return "query"


def simplify_vars(
    query_str: str, variables: dict[str, T.Any]
) -> (str, dict[str, T.Any]):
    new_vars = {}
    new_query_str = query_str + ";"
    for key, val in variables.items():
        if key in new_query_str:
            # strip key and add number to it
            stripped_key = key.split(RE_CODE)[0]
            i = 0
            while True:
                stripped_key_temp = stripped_key + f'{"" if i == 0 else i}'
                if stripped_key_temp not in new_vars:
                    stripped_key = stripped_key_temp
                    break
                i += 1
            new_vars[stripped_key] = val
            new_query_str = re.sub(
                rf"(\$)({key})(\W*)", r"\1" + stripped_key + r"\3", new_query_str
            )
    if new_query_str[-1] == ";":
        new_query_str = new_query_str[0:-1]
    return new_query_str, new_vars


async def query(
    client: edgedb.AsyncIOClient,
    query_str: str,
    variables: T.Optional[T.Dict[str, T.Any]] = None,
    only_one: bool = False,
    print_query: bool = True,
    print_variables: bool = False,
    print_raw_results: bool = False,
) -> T.Optional[dict]:
    """Returns a json str to be parsed by pydantic raw. Errors are raised by the lib!"""
    if not variables:
        variables = {}
    query_str, variables = simplify_vars(query_str, variables)
    query_func = client.query_json if not only_one else client.query_single_json
    start = time.time()
    try:
        with safe_span(
            op=f"edgedb.{operation_from_query_str(query_str)}",
            description=query_str[:200],
        ):
            j_str = await query_func(query=query_str, **variables)
        with safe_span(op="orjson.loads", description=f"len str: {len(j_str)}"):
            j = orjson.loads(j_str)
        if print_raw_results:
            debug(j)
    except edgedb.errors.ConstraintViolationError as e:
        print(f"{e=}")
        if "is prohibited by link target policy" in str(e):
            raise e
        if "violates exclusivity constraint" in str(e):
            field_name = str(e).split(" ")[0].replace("_", " ")
            raise ExecuteException(f"That {field_name} already exists in our system.")
        raise e
    except Exception as e:
        print(
            f"EdgeDB Query Exception: {e}, query_str and variables: {query_str=}, {variables=}"
        )
        raise e
    took_ms = (time.time() - start) * 1000
    print_s = ""
    if print_query:
        print_s += f" {query_str=} "
    if print_variables:
        print_s += f" {variables=} "
    if print_s:
        print_s = print_s.strip()
        print(print_s)
    print(f"took {took_ms}")
    return j



import typing as T
import re
from uuid import UUID
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo
import json
from fastapi.encoders import jsonable_encoder
import edgedb
from devtools import debug as dev_debug
from pydantic import BaseModel, PrivateAttr, Field

from .unset import UNSET
from .cache import Cache
from .execute import query as execute_query
from .batch import Batch
from .constants import random_str
from . import creating_strings
from .span import safe_span

InsertType = T.TypeVar("InsertType", bound=BaseModel)
PatchType = T.TypeVar("PatchType", bound=BaseModel)

ThisNodeType = T.TypeVar("ThisNodeType", bound="Node")

BaseModelType = T.TypeVar("BaseModelType", bound=BaseModel)

TempNodeType = T.TypeVar("TempNodeType", bound="Node")
if T.TYPE_CHECKING:
    from .resolver import Resolver

    TempResolverType = T.TypeVar("TempResolverType", bound=Resolver)
    ResolverType = T.TypeVar("ResolverType", bound=Resolver)


class NodeException(Exception):
    pass


def model_dict_to_str(
    *,
    d: dict,
    edgedb_conversion_map: T.Dict[str, T.Dict[str, str]],
    set_links_d: T.Dict[str, "Resolver"] = None,
    force_replace: bool = False,  # make this true for inserts
    set_strs: T.List[str] = None,
    error: bool = False,
) -> T.Tuple[str, T.Dict[str, T.Any]]:
    from .resolver import UpdateOperation

    field_strs: T.List[str] = []
    variables = {}
    for field_name, val in d.items():
        add_to_variables: bool = True
        type_cast_str = edgedb_conversion_map.get(field_name, {}).get("cast")
        if val is None:
            val = set()

        if not type_cast_str:
            # resolvers will come up here so ignore them
            if (set_links_d and field_name in set_links_d) or val == set():
                continue
            # if this is a resolver, it is fine... also what if it is none?
            if error:
                raise NodeException(
                    f"{field_name=}, {val=} not in {edgedb_conversion_map=}"
                )
            else:
                continue

        variable_name = f"{field_name}{random_str(10)}"
        # if enum, turn to str cast instead
        type_cast_str = re.sub(r"default::\w+", "std::str", type_cast_str)
        field_str = f"{field_name} := <{type_cast_str}>${variable_name}"
        if type_cast_str.endswith("::str") and (
            isinstance(val, dict) or isinstance(val, list)
        ):
            # for basemodel subfields that should go from dict to str
            val = json.dumps(jsonable_encoder(val))
        if type(val) is set:
            val = list(val)
            if "array<" not in type_cast_str:
                field_str = f"{field_name} := array_unpack(<array<{type_cast_str}>>${variable_name})"
                if len(val) == 0:
                    add_to_variables = False
                    field_str = f"{field_name} := {{}}"
        field_strs.append(field_str)
        if add_to_variables:
            variables[variable_name] = val
    if set_links_d:
        for field_name, resolver in set_links_d.items():
            update_operation_str = resolver.update_operation_str
            if force_replace:
                update_operation_str = UpdateOperation.REPLACE.value
            if not update_operation_str:
                raise NodeException(
                    f"Update resolver for {field_name} does not have an update operation."
                )
            field_str = (
                f"{field_name} {update_operation_str}"
                f" (SELECT DETACHED {resolver._model_name}{resolver.all_filters_str()})"
            )
            field_strs.append(field_str)
            for variable_name, variable_value in resolver._query_variables.items():
                if variable_name in variables:
                    raise NodeException(
                        f"Query variable {variable_name} for update resolver {field_name} "
                        f"already exists in variables {variables}."
                    )
                variables[variable_name] = variable_value
    if set_strs:
        field_strs.extend(set_strs)
    s = f'{{ {", ".join(field_strs)} }}'
    return s, variables


def make_set_links_d(
    model: BaseModel, model_d: dict, default_replace: bool = False
) -> T.Dict[str, "ResolverType"]:
    """mutates model_d and returns set links d"""
    from .resolver import Resolver, UpdateOperation

    set_links_d = {}
    for field_name in model.__fields__.keys():
        val = getattr(model, field_name)
        if isinstance(val, Resolver):
            if not val._update_operation and default_replace:
                val.update_operation(UpdateOperation.REPLACE)
            set_links_d[field_name] = val
            if field_name in model_d:
                del model_d[field_name]
    return set_links_d


def basemodel_to_str(
    model: BaseModelType,
    default_replace: bool = False,
    fields_to_exclude: T.Set[str] = None,
    force_replace: bool = False,
) -> T.Tuple[str, T.Dict[str, T.Any]]:
    # should exclude_none = True?
    # model_d = model.dict(exclude_none=True, exclude=fields_to_exclude)
    # TODO is this right?
    model_d = model.dict(exclude=fields_to_exclude)

    set_links_d = make_set_links_d(
        model=model, model_d=model_d, default_replace=default_replace
    )

    return model_dict_to_str(
        d=model_d,
        edgedb_conversion_map=model._edgedb_conversion_map,
        set_links_d=set_links_d,
        force_replace=force_replace,
    )


class Node(BaseModel, T.Generic[InsertType, PatchType]):
    id: UUID = Field(..., allow_mutation=False)

    _cache: Cache = PrivateAttr(default_factory=Cache)
    _used_resolver: "ResolverType" = PrivateAttr(None)
    _original_dict: dict = PrivateAttr(None)
    _deleted: bool = PrivateAttr(None)

    _edgedb_conversion_map: T.ClassVar[T.Dict[str, T.Dict[str, str]]]
    _link_conversion_map: T.ClassVar[T.Dict[str, T.Dict[str, str]]]
    _computed_properties: T.ClassVar[T.Set[str]]
    _appendix_properties: T.ClassVar[T.Set[str]]
    _basemodel_properties: T.ClassVar[T.Set[str]]

    _extra: dict = PrivateAttr(default=dict())

    __updated_fields__: T.Set[str] = PrivateAttr(default=set())

    def __setattr__(self, key, value):
        self.__updated_fields__.add(key)
        if key in self._appendix_properties:
            # TODO is this the desired effect?
            #  Probably, if you set an appendix field, add it to fields returned in future
            self._used_resolver.include_fields(key)
            k = f"{key}_"
            if k in self.__fields__.keys():
                key = k
        return super().__setattr__(key, value)

    def updated_fields(self, *updated_fields: str) -> None:
        self.__updated_fields__.update(updated_fields)

    def _clear_updated_fields(self) -> None:
        self.__updated_fields__ = set()

    @property
    def extra(self) -> dict:
        return self._extra

    class Config:
        validate_assignment = True
        underscore_attrs_are_private = False

    class GraphORM:
        model_name: str = None
        client: edgedb.AsyncIOClient
        exclusive_fields: T.Set[str] = None
        resolver_type: T.ClassVar[T.Type["ResolverType"]]
        updatable_fields: T.ClassVar[T.Set[str]] = set()

    @classmethod
    def get_aliases(cls) -> T.Set[str]:
        return {field.alias for field in cls.__fields__.values()}

    def __eq__(self, other: TempNodeType) -> bool:
        if not isinstance(other, Node):
            return False
        return f"{hash(self)}{self.json()}" == f"{hash(other)}{other.json()}"

    def __hash__(self) -> int:
        return hash(f"{self.__class__.__name__}{self.json()}")

    def to_resolver(self) -> "ResolverType":
        return self._used_resolver.__class__().filter(
            filter_str=".id = <uuid>$id", variables={"id": self.id}
        )

    @classmethod
    def to_resolver_many(
        cls: T.Type[ThisNodeType], nodes: T.List[ThisNodeType]
    ) -> "ResolverType":
        return cls.GraphORM.resolver_type().filter(
            ".id in array_unpack(<array<uuid>>$ids)",
            variables={"ids": [n.id for n in nodes]},
        )

    @property
    def cache(self) -> Cache:
        return self._cache

    @classmethod
    @property
    def db_client(cls) -> edgedb.AsyncIOClient:
        return cls.GraphORM.client

    def __repr__(self) -> str:
        r = super().__repr__()
        r = f"{r}, cache: {repr(self.cache)}" if not self.cache.is_empty() else r
        return r

    """CACHE"""

    async def resolve(
        self,
        edge_name: str,
        edge_resolver: "TempResolverType",
        refresh: bool = False,
        revert_to_first: bool = False,
        allow_fetching: bool = True,
        debug: bool = False,
    ) -> T.Optional[T.Union[TempNodeType, T.List[TempNodeType]]]:
        if refresh:
            # FIXME only remove specific resolver? But what about subsets
            self.cache.clear(edge=edge_name)
        val = self.cache.val(
            edge=edge_name, resolver=edge_resolver, revert_to_first=revert_to_first
        )
        if val is not UNSET:
            return val
        # if unset, must get this
        if not allow_fetching:
            raise NodeException(f"Not allowed to fetch edge {edge_name}.")
        self.used_resolver._nested_resolvers.add(edge=edge_name, resolver=edge_resolver)
        self._used_resolver.clear_top_level_filters_and_variables()
        obj: Node = await self.used_resolver.gerror(id=self.id)
        self.cache.replace(edge=edge_name, cache_nodes=obj.cache.get(edge_name))
        return self.cache.val(edge=edge_name, resolver=edge_resolver)

    """CRUDS"""

    @property
    def used_resolver(self) -> "ResolverType":
        return self._used_resolver

    @staticmethod
    def validate_upsert_fields(
        upsert_given_conflict_on: str = None,
        return_conflicting_model_on: str = None,
        custom_on_conflict_str: str = None,
    ) -> None:
        if custom_on_conflict_str:
            if return_conflicting_model_on or upsert_given_conflict_on:
                raise NodeException(
                    "You cannot give a custom conflict string with other conflict inputs."
                )
        if return_conflicting_model_on and upsert_given_conflict_on:
            raise NodeException(
                "You cannot both suppress a conflict and upsert given a conflict."
            )

    def build_insert_str(
        self,
        use_all_fields: bool = False,  # true for FOR cause some objects might have null values while some do not
        upsert_given_conflict_on: str = None,
        return_conflicting_model_on: str = None,
        custom_on_conflict_str: str = None,
    ) -> str:
        ...

    @classmethod
    async def add(
        cls: T.Type[ThisNodeType],
        insert: InsertType,
        given_resolver: "ResolverType" = None,
        batch: Batch = None,
        given_client: edgedb.AsyncIOClient = None,
        upsert_given_conflict_on: str = None,
        return_conflicting_model_on: str = None,
        custom_on_conflict_str: str = None,
    ) -> T.Optional[ThisNodeType]:
        """Turn insert into a string with variables"""
        if not given_resolver:
            given_resolver = cls.GraphORM.resolver_type()
        cls.validate_upsert_fields(
            upsert_given_conflict_on=upsert_given_conflict_on,
            return_conflicting_model_on=return_conflicting_model_on,
            custom_on_conflict_str=custom_on_conflict_str,
        )
        # I think default_replace=True is correct here. Your intention is to make sure these links are SET, not updated
        # if your intention is for them to update on upsert, then add ADD to the resolver
        s, variables = basemodel_to_str(
            insert, default_replace=True, force_replace=True
        )
        insert_str = f"INSERT {cls._model_name} {s}"
        conflict_str = ""
        if custom_on_conflict_str:
            conflict_str = custom_on_conflict_str
        elif upsert_given_conflict_on:
            # replace s and variables
            # insert_dict = insert.dict(exclude_none=True)
            # TODO again for exclude_none is this right?
            insert_dict = insert.dict()
            fields_to_exclude = set(insert_dict.keys()) - cls.GraphORM.updatable_fields
            update_s, update_variables = basemodel_to_str(
                insert, default_replace=True, fields_to_exclude=fields_to_exclude
            )
            variables.update(update_variables)
            conflict_str = f"UNLESS CONFLICT ON .{upsert_given_conflict_on} ELSE (UPDATE {cls._model_name} SET {update_s})"
        elif return_conflicting_model_on:
            conflict_str = f"UNLESS CONFLICT ON .{return_conflicting_model_on} ELSE (SELECT {cls._model_name})"
        if conflict_str:
            insert_str += f" {conflict_str}"
        if batch:
            batch.add(line=insert_str, variables={**variables})
            return None
        select_str = f"SELECT model {given_resolver.to_str()}"
        add_s = f"WITH model := ({insert_str}) {select_str}"
        with safe_span(op=f"edgedb.add.{cls._model_name}"):
            raw_d = await execute_query(
                client=given_client or cls.db_client,
                query_str=add_s,
                variables=variables,
                only_one=True,
            )
        with safe_span(op=f"parse.{cls._model_name}"):
            return given_resolver.parse_nested_obj(raw_d)

    @classmethod
    async def add_many(
        cls: T.Type[ThisNodeType],
        inserts: T.List[InsertType],
        edge_filter_strs: T.List[str] = None,
        given_resolver: "ResolverType" = None,
        given_client: edgedb.AsyncIOClient = None,
        upsert_given_conflict_on: str = None,
        return_conflicting_model_on: str = None,
        custom_on_conflict_str: str = None,
    ) -> T.List[ThisNodeType]:
        """This returns in no specific order"""
        if not given_resolver:
            given_resolver = cls.GraphORM.resolver_type()
        if not edge_filter_strs:
            edge_filter_strs = []
        cls.validate_upsert_fields(
            upsert_given_conflict_on=upsert_given_conflict_on,
            return_conflicting_model_on=return_conflicting_model_on,
            custom_on_conflict_str=custom_on_conflict_str,
        )
        if not inserts:
            return list()
        insert_str = creating_strings.insert_str_from_cls(
            insert_cls=inserts[0].__class__,
            node_cls=cls,
            edge_filter_strs=edge_filter_strs,
            upsert_given_conflict_on=upsert_given_conflict_on,
            return_conflicting_model_on=return_conflicting_model_on,
            custom_on_conflict_str=custom_on_conflict_str,
        )
        full_insert_str = f"""
with
    raw_data := <json>$data,
for item in json_array_unpack(raw_data) union ({insert_str}) {given_resolver.to_str()}
        """
        print(f"{full_insert_str=}")
        data: T.List[T.Dict[str, T.Any]] = []
        for insert in inserts:
            d = insert.dict(exclude_none=True)
            # TODO for bulk adding, best to exclude
            # d = insert.dict()
            # if the cast says they should be strings and they are not, turn them into json
            for field_name, value in d.items():
                if cast := cls._edgedb_conversion_map.get(field_name, {}).get("cast"):
                    if (
                        cast.endswith("::str")
                        and not isinstance(value, str)
                        and not type(value) in [set]
                    ):
                        # if ITERABLE of strs, ignore. For now just SET, cause what if list of objs
                        d[field_name] = json.dumps(jsonable_encoder(value))
            # get fields that are LINKS
            for link_name in cls._link_conversion_map.keys():
                if link_resolver := getattr(insert, link_name, None):
                    d[link_name] = {**link_resolver._query_variables}
            data.append(d)
        data_json = json.dumps(jsonable_encoder(data))
        len_inserts = len(inserts)
        with safe_span(
            op=f"edgedb.add_many.{cls._model_name}", description=f"{len_inserts=}"
        ):
            raw_d = await execute_query(
                client=given_client or cls.db_client,
                query_str=full_insert_str,
                variables={"data": data_json},
            )
        with safe_span(op=f"parse.{cls._model_name}", description=f"{raw_d=}"):
            return [given_resolver.parse_nested_obj(d) for d in raw_d]

    @classmethod
    async def add_many_old(
        cls: T.Type[ThisNodeType],
        inserts: T.List[InsertType],
        given_resolver: "ResolverType" = None,
        given_client: edgedb.AsyncIOClient = None,
    ) -> T.List[ThisNodeType]:
        """This appears to have a max limit of something between 30 and 50, so must break it up"""
        if not given_resolver:
            given_resolver = cls.GraphORM.resolver_type()
        insert_strs: T.List[str] = []
        variables = {}
        for i, insert in enumerate(inserts):
            insert_str, insert_variables = basemodel_to_str(model=insert)
            insert_strs.append(insert_str)
            variables.update(insert_variables)
        insert_lines: T.List[str] = []
        model_name = cls._model_name
        model_enumerated_names: T.List[str] = []
        for i, insert_str in enumerate(insert_strs):
            model_enumerated_name = f"model_{i}"
            insert_lines.append(
                f"{model_enumerated_name} := (INSERT {model_name} {insert_str})"
            )
            model_enumerated_names.append(model_enumerated_name)
        insert_lines.append(f'models := ({" UNION ".join(model_enumerated_names)})')
        add_many_s = (
            f'WITH {", ".join(insert_lines)} SELECT models {given_resolver.to_str()}'
        )
        raw_d = await execute_query(
            client=given_client or cls.db_client,
            query_str=add_many_s,
            variables=variables,
        )
        return [given_resolver.parse_nested_obj(d) for d in raw_d]

    async def delete(
        self,
        given_resolver: "ResolverType" = None,
        batch: Batch = None,
        given_client: edgedb.AsyncIOClient = None,
    ) -> None:
        if not given_resolver:
            given_resolver = self._used_resolver
        given_resolver.clear_top_level_filters_and_variables()
        given_resolver.filter(filter_str=".id = <uuid>$id", variables={"id": self.id})
        if batch:
            batch.add(
                line=given_resolver.inner_delete_str(),
                variables=given_resolver._get_nested_query_variables(),
            )
            return
        with safe_span(op=f"edgedb.delete.{self._model_name}"):
            raw_d = await execute_query(
                client=given_client or self.db_client,
                query_str=given_resolver.full_delete_str(),
                variables=given_resolver._get_nested_query_variables(),
                only_one=True,
            )
        if not raw_d:
            raise NodeException(
                f"No delete for {self._model_name}: {self.id} was registered."
            )
        with safe_span(op=f"parse.{self._model_name}"):
            node = given_resolver.parse_nested_obj(raw_d)
        node._deleted = True
        self.hydrate(new_node=node)

    @classmethod
    async def delete_many(
        cls: T.Type[ThisNodeType],
        models: T.List[ThisNodeType],
        given_resolver: "ResolverType" = None,
    ) -> None:
        if not given_resolver:
            given_resolver = cls.GraphORM.resolver_type()
        given_resolver.clear_top_level_filters_and_variables()
        given_resolver.filter(
            filter_str=".id in array_unpack(<array<uuid>>$ids)",
            variables={"ids": [model.id for model in models]},
        )
        len_models = len(models)
        with safe_span(
            op=f"edgedb.delete_many.{cls._model_name}", description=f"{len_models=}"
        ):
            raw_d = await execute_query(
                client=cls.db_client,
                query_str=given_resolver.full_delete_str(),
                variables=given_resolver._get_nested_query_variables(),
            )
        with safe_span(op=f"parse.{cls._model_name}", description=f"{len(raw_d)=}"):
            new_nodes = [given_resolver.parse_nested_obj(d) for d in raw_d]
        models_by_id: T.Dict[str, ThisNodeType] = {model.id: model for model in models}
        for new_node in new_nodes:
            new_node._deleted = True
            models_by_id[new_node.id].hydrate(new_node=new_node)

    @classmethod
    async def delete_from_filtered_resolver(
        cls,
        filtered_resolver: "ResolverType",
        batch: Batch = None,
        given_client: edgedb.AsyncIOClient = None,
        only_one: bool = False,
    ) -> T.Optional[T.List[ThisNodeType]]:
        if batch:
            batch.add(
                line=filtered_resolver.inner_delete_str(),
                variables=filtered_resolver._get_nested_query_variables(),
            )
            return
        with safe_span(op=f"edgedb.delete_from_filtered_resolver.{cls._model_name}"):
            raw_d = await execute_query(
                client=given_client or cls.db_client,
                query_str=filtered_resolver.full_delete_str(),
                variables=filtered_resolver._get_nested_query_variables(),
                only_one=only_one,
            )

        def parse_deleted(inner_d: dict) -> ThisNodeType:
            node = filtered_resolver.parse_nested_obj(inner_d)
            node._deleted = True
            return node

        with safe_span(
            op=f"parse.{cls._model_name}",
            description=f"only_one:{only_one}, {len(raw_d)=}",
        ):
            if only_one:
                return parse_deleted(raw_d)
            else:
                return [parse_deleted(d) for d in raw_d]

    @classmethod
    def build_update_d_from_patch(cls, patch: PatchType) -> dict:
        update_d = {}
        for field in patch.__fields__.keys():
            val = getattr(patch, field)
            if val is not UNSET:
                if val is None:
                    val = set()
                elif isinstance(val, BaseModel):
                    val = val.json()
                update_d[field] = val
        return update_d

    def build_update_d(self) -> dict:
        fields_to_update = self.__updated_fields__ & self.GraphORM.updatable_fields
        update_d = {}
        for field in fields_to_update:
            val = getattr(self, field)
            if val is None:
                val = set()
            elif isinstance(val, BaseModel):
                val = val.json()
            update_d[field] = val
        return update_d

    def get_update_d_old(self) -> dict:
        """DEP. Keep incase new one ^'build_update_d' is not good"""
        # for appendix properties
        # if you want to edit an appendix field, you must request it
        current_d = {**self.extra, **self.dict()}
        for field, val in current_d.items():
            print(f"{field=}, {val=}")
            if (
                field in self._basemodel_properties
                and field in self.GraphORM.updatable_fields
                and isinstance(getattr(self, field, None), BaseModel)
            ):
                current_d[field] = getattr(self, field).json()
        original_d = self._original_dict
        unset_key = "_UNSET_12398120"
        update_d = {}
        for field_name in self.GraphORM.updatable_fields:
            current_val = current_d.get(field_name, unset_key)
            original_val = original_d.get(field_name, unset_key)
            if current_val != original_val:
                if current_val != unset_key:
                    # if current val is not unset -> always set
                    print(f"{current_val=}, {original_val=}")
                    if current_val is not None:
                        if (
                            isinstance(current_val, datetime)
                            and current_val.isoformat() == original_val
                        ):
                            continue
                        else:
                            if isinstance(current_val, datetime):
                                print(
                                    f"datetimes diff for update: {current_val.isoformat()=}, {original_val=}"
                                )
                            update_d[field_name] = current_val
                    else:
                        update_d[field_name] = set()
                else:
                    # current val is UNSET but original val is not
                    update_d[field_name] = set()
        return update_d

    @classmethod
    async def _update(
        cls: T.Type[ThisNodeType],
        resolver: "ResolverType",
        update_d: dict,
        set_strs: T.List[str] = None,
        set_links_d: T.Dict[str, "Resolver"] = None,
        batch: Batch = None,
        given_client: edgedb.AsyncIOClient = None,
        only_one: bool = True,
    ) -> T.Optional[T.Union[ThisNodeType, T.List[ThisNodeType]]]:
        s, variables = model_dict_to_str(
            d=update_d,
            edgedb_conversion_map=cls._edgedb_conversion_map,
            set_links_d=set_links_d,
            set_strs=set_strs,
        )

        resolver.validate_against_query_variables(variables)
        update_inner_str = f"UPDATE {cls._model_name} {resolver._filter_str} SET {s}"

        variables_to_use = {**resolver._get_nested_query_variables(), **variables}

        if batch:
            batch.add(line=update_inner_str, variables=variables_to_use)
            return None

        update_s = (
            f"WITH model := ({update_inner_str}) "
            f"SELECT model {resolver.to_str(include_filters=False)}"
        )

        with safe_span(op=f"edgedb.update.{cls._model_name}"):
            raw_d = await execute_query(
                client=given_client or cls.db_client,
                query_str=update_s,
                variables=variables_to_use,
                only_one=only_one,
            )
        if not raw_d:
            raise NodeException(f"No update for {cls._model_name} was registered.")
        with safe_span(op=f"parse.{cls._model_name}"):
            if only_one:
                return resolver.parse_nested_obj(raw_d)
            else:
                return [resolver.parse_nested_obj(d) for d in raw_d]

    @classmethod
    async def update_from_patch(
        cls: T.Type[ThisNodeType],
        patch: PatchType,
        resolver_with_filter: "ResolverType",
        set_strs: T.List[str] = None,
        only_one: bool = True,
        update_last_updated: bool = False,
        batch: Batch = None,
        given_client: edgedb.AsyncIOClient = None,
    ) -> T.Optional[T.Union[ThisNodeType, T.List[ThisNodeType]]]:
        if update_last_updated:
            if getattr(patch, "last_updated", None) is UNSET:
                patch.last_updated = datetime.now(tz=ZoneInfo("America/New_York"))
        update_d = cls.build_update_d_from_patch(patch=patch)
        set_links_d = make_set_links_d(model=patch, model_d=update_d)
        if not update_d and not set_links_d:
            raise NodeException("No update registered.")
        node_or_nodes = await cls._update(
            resolver=resolver_with_filter,
            update_d=update_d,
            set_strs=set_strs,
            set_links_d=set_links_d,
            batch=batch,
            given_client=given_client,
            only_one=only_one,
        )
        if batch:
            return None
        return node_or_nodes

    async def update(
        self,
        given_resolver: "ResolverType" = None,
        error_if_no_update: bool = False,
        set_links_d: T.Dict[str, "Resolver"] = None,
        batch: Batch = None,
        given_client: edgedb.AsyncIOClient = None,
    ) -> None:
        """set_links_d -> {people_i_follow += (select Person filter .first_name = <str>first_name)}"""
        if not given_resolver:
            given_resolver = self._used_resolver
        update_d = self.build_update_d()
        if not update_d:
            message = f"No update registered for {self.id=}."
            if error_if_no_update:
                raise NodeException(message)
            else:
                print(message)
            return

        given_resolver.clear_top_level_filters_and_variables()
        given_resolver.filter(filter_str=".id = <uuid>$id", variables={"id": self.id})
        node = await self._update(
            resolver=given_resolver,
            update_d=update_d,
            set_links_d=set_links_d,
            batch=batch,
            given_client=given_client,
        )
        if batch:
            return None
        self.hydrate(new_node=node)

    @classmethod
    @property
    def _model_name(cls) -> str:
        model_name = getattr(cls.GraphORM, "model_name", None)
        if model_name is None:
            model_name = cls.__name__
        return model_name

    @classmethod
    @property
    def _exclusive_fields(cls) -> T.Set[str]:
        exclusive_fields = getattr(cls.GraphORM, "exclusive_fields", set())
        exclusive_fields.add("id")
        return exclusive_fields

    def hydrate(
        self, new_node: ThisNodeType, clear_updated_fields: bool = True
    ) -> None:
        """Turns this node into the new node"""
        if clear_updated_fields:
            self._clear_updated_fields()
        for field_name in new_node.__fields__.keys():
            new_field = getattr(new_node, field_name)
            old_field = getattr(self, field_name)
            if new_field != old_field:
                setattr(self, field_name, getattr(new_node, field_name))
        for private_attr_name in new_node.__private_attributes__.keys():
            setattr(self, private_attr_name, getattr(new_node, private_attr_name))

    async def refresh(self: ThisNodeType) -> None:
        if self._deleted:
            raise NodeException("Node is already deleted.")
        self._used_resolver.clear_top_level_filters_and_variables()
        new_node = await self._used_resolver.get(id=self.id)
        self.hydrate(new_node=new_node)



import typing as T
import random
import string

ALIAS_PATTERN = r">\$(\w+)"
RE_CODE = "04730"


def random_str(n: int, include_re_code: bool = True) -> str:
    pre_str = RE_CODE if include_re_code else ""
    return pre_str + "".join(random.choices(string.ascii_letters + string.digits, k=n))


def chunk_list(lst: list, chunk_size: int) -> T.List[list]:
    return [lst[i : i + chunk_size] for i in range(0, len(lst), chunk_size)]


def flatten_list(lst: T.List[list]) -> list:
    return [j for sub in lst for j in sub]



from .node import Node, NodeException
from .resolver import Resolver, ResolverException, UpdateOperation, FilterConnector
from .batch import Batch
from .unset import (
    UnsetType,
    ComputedPropertyException,
    AppendixPropertyException,
    UNSET,
)
from .validators import from_str, enum_from_str
from .enums import PropertyCardinality
from .base_patch import BasePatch



import typing as T
import re
import asyncio
from edgedb import AsyncIOClient
from pydantic import BaseModel
from .constants import ALIAS_PATTERN, random_str, chunk_list, flatten_list
from .execute import query as execute_query


class BatchException(Exception):
    pass


class Batch(BaseModel):
    client: AsyncIOClient
    print_status: bool = True
    debug: bool = False
    mutation_lines: T.List[str] = []
    query_variables: T.Dict[str, T.Any] = {}

    class Config:
        arbitrary_types_allowed = True

    def add(self, line: str, variables: T.Dict[str, T.Any] = None) -> None:
        """for add, update, delete"""
        aliases = re.findall(ALIAS_PATTERN, line)
        for alias in aliases:
            # find all instances of that alias, change the variable
            random_model_id = random_str(10)
            new_alias = f"{alias}{random_model_id}"
            if new_alias in self.query_variables:
                raise BatchException(f"Found a duplicate new alias {new_alias}")
            if alias in variables:
                variables[new_alias] = variables[alias]
                del variables[alias]
                pattern_to_sub = fr"(>\$)({alias})(\W*)"
                line = re.sub(pattern_to_sub, fr"\1{new_alias}\3", line)

        self.mutation_lines.append(line)
        self.query_variables.update(variables)

    async def commit_chunk(
        self, mutation_lines_chunk: T.List[str], chunk_id: int
    ) -> T.List[str]:
        lines_strs: T.List[str] = []
        model_name_strs: T.List[str] = []
        for i, line in enumerate(mutation_lines_chunk):
            model_name = f"model{i}"
            line_str = f"{model_name} := ({line})"
            model_name_strs.append(model_name)
            lines_strs.append(line_str)
        lines_strs.append(f'models := {{{", ".join(model_name_strs)}}}')
        s = f"WITH {','.join(lines_strs)} SELECT models {{ id }};"
        # now remove the non-used aliases from variables
        used_aliases = re.findall(ALIAS_PATTERN, str(lines_strs))
        valid_variables = {
            key: val for key, val in self.query_variables.items() if key in used_aliases
        }
        raw_d = await execute_query(
            client=self.client,
            query_str=s,
            variables=valid_variables,
            print_query=False,
        )
        ids = [d["id"] for d in raw_d]
        if self.print_status:
            print(f"executed {chunk_id=}, {ids=}")
            if self.debug:
                print(f"QUERY STRING: {s=}")
        return ids

    async def commit(self, chunk_size: int = None) -> T.List[str]:
        """returns a list of ids from the nodes touched.
        It is atomic so if you have a chunk size it will lose the atomic nature.
        But also it will fail if it has over ~20 objects"""
        if chunk_size:
            chunks = chunk_list(lst=self.mutation_lines, chunk_size=chunk_size)
        else:
            chunks = [self.mutation_lines]
        proms = [
            self.commit_chunk(mutation_lines_chunk=chunk, chunk_id=i)
            for i, chunk in enumerate(chunks)
        ]
        ids_lst = await asyncio.gather(*proms)
        # should i check for errors?
        return flatten_list(lst=ids_lst)



import typing as T
import re
from enum import Enum
import edgedb
from devtools import debug
from pydantic import BaseModel, Field, parse_obj_as
from .execute import query
from .enums import PropertyCardinality


class IntrospectionException(Exception):
    pass


edgedb_to_python_type_mapping = {
    "str": "str",
    "bool": "bool",
    "int16": "int",
    "int32": "int",
    "int64": "int",
    "float32": "float",
    "float64": "float",
    "bigint": "float",
    "decimal": "Decimal",
    "json": "str",
    "uuid": "UUID",
    "bytes": "bytes",
    "datetime": "datetime",
    "duration": "timedelta",
    "local_datetime": "datetime",
    "local_date": "date",
    "local_time": "time",
    "relative_duration": "RelativeDuration",
    "sequence": "int",
}


class Cardinality(str, Enum):
    One = "One"
    Many = "Many"


class ElementType(BaseModel):
    name: str

    @property
    def type_name(self) -> str:
        return self.name.split("::")[-1]


class Target(BaseModel):
    name: str
    # prob do not need element type
    element_type: ElementType = None

    @property
    def model_name(self) -> str:
        return self.name.split("::")[-1]


class Annotation(BaseModel):
    name: str
    value: str = Field(..., alias="@value")

    @property
    def not_insertable(self) -> bool:
        return "not_insertable" in self.name and bool(self.value)


class Constraint(BaseModel):
    name: str

    @property
    def is_exclusive(self) -> bool:
        return self.name == "std::exclusive"


class Index(BaseModel):
    expr: str

    @property
    def fields(self) -> T.List[str]:
        pattern = r"\.(\w+)"
        return list(re.findall(pattern, self.expr))


class ParamType(str, Enum):
    LINK = "LINK"
    PROPERTY = "PROPERTY"


class Param(BaseModel):
    name: str
    default: T.Optional[str] = None
    cardinality: Cardinality
    required: bool
    target: Target
    readonly: bool

    constraints: T.List[Constraint]
    annotations: T.List[Annotation]

    expr: T.Optional[str] = None

    @property
    def not_insertable(self) -> bool:
        if not self.annotations:
            return False
        for annotation in self.annotations:
            if annotation.not_insertable:
                return True
        return False

    @property
    def is_computed(self) -> bool:
        return bool(self.expr)

    @property
    def is_exclusive(self) -> bool:
        for constraint in self.constraints:
            if constraint.is_exclusive:
                return True
        return False

    def edgedb_core_type(self) -> str:
        pattern = r"::(\w+)"
        core_types = re.findall(pattern, self.target.name)
        if len(core_types) != 1:
            raise IntrospectionException(
                f"Problem parsing core types {core_types=} for {self.name=}, {self.dict()=}"
            )
        return core_types[0]

    @property
    def type_str(self) -> str:
        """returns [Person] or T.Optional[Person] or T.List[str] or str"""
        # if it is a property, make MANY a SET. If it is a link, make it a List
        edgedb_type_str = self.target.name
        edgedb_core_type = self.edgedb_core_type()

        python_type = (
            edgedb_to_python_type_mapping.get(edgedb_core_type) or edgedb_core_type
        )
        s = python_type
        if edgedb_type_str.startswith("array"):
            s = f"T.List[{s}]"
        elif edgedb_type_str.startswith("tuple"):
            s = f"T.Tuple[{s}]"
        if self.cardinality == Cardinality.Many:
            iterable = "Set" if self.__class__.__name__ == "Property" else "List"
            s = f"T.{iterable}[{s}]"
        if self.required is False:
            s = f"T.Optional[{s}]"
        return s

    def type_str_basemodel(
        self, module_name: str, cardinality: PropertyCardinality
    ) -> str:
        type_str = self.type_str
        if cardinality == PropertyCardinality.MANY:
            module_name = f"T.List[{module_name}]"
        if "[" in type_str:
            type_str = re.sub(r"\[(\w+)\]", f"[{module_name}]", type_str)
        else:
            type_str = module_name
        return type_str


class Link(Param):
    pass


class Property(Param):
    pass


class ObjectType(BaseModel):
    name: str
    links: T.List[Link]
    properties: T.List[Property]
    constraints: T.List[Constraint]
    indexes: T.List[Index]

    @property
    def node_name(self) -> str:
        return self.name.split("::")[-1]


object_types_query_str = """
with module schema
select ObjectType {
    name,
    links: {
        name,
        default,
        cardinality,
        required,
        expr,
        target: {
            name,
            [is Array].element_type: {
                name
            }
        },
        constraints: {
            name
        },
        readonly,
        annotations: {
            name,
            @value
        },
    },
    properties: {
        name,
        default,
        cardinality,
        required,
        expr,
        target: {
            name,
            [is Array].element_type: {
                name
            }
        },
        constraints: {
            name
        },
        readonly,
        annotations: {
            name,
            @value
        },
    },
    constraints: {
        name
    },
    indexes: {
        expr
    }
} filter contains(.name, 'default::');
"""


async def introspect_objects(client: edgedb.AsyncIOClient) -> T.List[ObjectType]:
    object_types_raw = await query(client, object_types_query_str)
    object_types = parse_obj_as(T.List[ObjectType], object_types_raw)
    return object_types


scalar_types_query_str = """
with module schema
select ScalarType {
    name,
    enum_values,
    abstract,
    annotations: {
        name,
        @value
    },
    constraints: {
        name
    }
} filter contains(.name, 'default::');
"""


class ScalarType(BaseModel):
    name: str
    enum_values: T.List[str]
    abstract: bool
    annotations: T.List[Annotation]
    constraints: T.List[Constraint]

    @property
    def node_name(self) -> str:
        return self.name.split("::")[-1]


async def introspect_scalars(client: edgedb.AsyncIOClient) -> T.List[ScalarType]:
    scalar_types_raw = await query(client, scalar_types_query_str)
    scalar_types = parse_obj_as(T.List[ScalarType], scalar_types_raw)
    return scalar_types
